import argparse
import sys
from argparse import Namespace

from redisvl import __version__
from redisvl.cli.log import get_logger

logger = get_logger("[RedisVL]")


class Version:
    usage = "\n".join(
        [
            "rvl version [<args>]\n",
            "\n",
        ]
    )

    def __init__(self):
        parser = argparse.ArgumentParser(usage=self.usage)
        parser.add_argument(
            "-s", "--short", help="Print only the version number", action="store_true"
        )

        args = parser.parse_args(sys.argv[2:])
        self.version(args)

    def version(self, args: Namespace):
        if args.short:
            print(__version__)
        else:
            logger.info(f"RedisVL version {__version__}")



from typing import Callable, Dict, List, Optional

from redisvl.vectorize.base import BaseVectorizer


class HFTextVectorizer(BaseVectorizer):
    # TODO - add docstring
    def __init__(self, model: str, api_config: Optional[Dict] = None):
        super().__init__(model)
        try:
            from sentence_transformers import SentenceTransformer
        except ImportError:
            raise ImportError(
                "HFTextVectorizer requires sentence-transformers library. Please install with pip install sentence-transformers"
            )

        self._model_client = SentenceTransformer(self._model)

        try:
            self._dims = self._set_model_dims()
        except:
            raise ValueError("Error setting embedding model dimensions")

    def _set_model_dims(self):
        embedding = self._model_client.encode(["dimension check"])[0]
        return len(embedding)

    def embed(
        self,
        text: str,
        preprocess: Optional[Callable] = None,
        as_buffer: Optional[float] = False,
    ) -> List[float]:
        """Embed a chunk of text using the Hugging Face sentence transformer.

        Args:
            text (str): Chunk of text to embed.
            preprocess (Optional[Callable], optional): Optional preprocessing callable to
                perform before vectorization. Defaults to None.
            as_buffer (Optional[float], optional): Whether to convert the raw embedding
                to a byte string. Defaults to False.

        Returns:
            List[float]: Embedding.

        Raises:
            TypeError: If the wrong input type is passed in for the text.
        """
        if not isinstance(text, str):
            raise TypeError("Must pass in a str value to embed.")

        if preprocess:
            text = preprocess(text)
        embedding = self._model_client.encode([text])[0]
        return self._process_embedding(embedding.tolist(), as_buffer)

    def embed_many(
        self,
        texts: List[str],
        preprocess: Optional[Callable] = None,
        batch_size: int = 1000,
        as_buffer: Optional[float] = None,
    ) -> List[List[float]]:
        """Asynchronously embed many chunks of texts using the Hugging Face sentence
        transformer.

        Args:
            texts (List[str]): List of text chunks to embed.
            preprocess (Optional[Callable], optional): Optional preprocessing callable to
                perform before vectorization. Defaults to None.
            batch_size (int, optional): Batch size of texts to use when creating
                embeddings. Defaults to 10.
            as_buffer (Optional[float], optional): Whether to convert the raw embedding
                to a byte string. Defaults to False.

        Returns:
            List[List[float]]: List of embeddings.

        Raises:
            TypeError: If the wrong input type is passed in for the test.
        """
        if not isinstance(texts, list):
                raise TypeError("Must pass in a list of str values to embed.")
        if  len(texts) > 0 and not isinstance(texts[0], str):
                raise TypeError("Must pass in a list of str values to embed.")

        embeddings: List = []
        for batch in self.batchify(texts, batch_size, preprocess):
            batch_embeddings = self._model_client.encode(batch)
            embeddings.extend(
                [
                    self._process_embedding(embedding.tolist(), as_buffer)
                    for embedding in batch_embeddings
                ]
            )
        return embeddings



from typing import Callable, Dict, List, Optional

from tenacity import retry, stop_after_attempt, wait_random_exponential
from tenacity.retry import retry_if_not_exception_type

from redisvl.vectorize.base import BaseVectorizer


class OpenAITextVectorizer(BaseVectorizer):
    """OpenAI text vectorizer

    This vectorizer uses the OpenAI API to create embeddings for text. It requires an
    API key to be passed in the api_config dictionary. The API key can be obtained from
    https://api.openai.com/.
    """
    def __init__(self, model: str, api_config: Optional[Dict] = None):
        """Initialize the OpenAI vectorizer.

        Args:
            model (str): Model to use for embedding.
            api_config (Optional[Dict], optional): Dictionary containing the API key.
                Defaults to None.

        Raises:
            ImportError: If the openai library is not installed.
            ValueError: If the API key is not provided.
        """
        super().__init__(model)
        # Dynamic import of the openai module
        try:
            global openai
            import openai
        except ImportError:
            raise ImportError(
                "OpenAI vectorizer requires the openai library. Please install with pip install openai"
            )

        if not api_config or "api_key" not in api_config:
            raise ValueError("OpenAI API key is required in api_config")

        openai.api_key = api_config["api_key"]
        self._model_client = openai.Embedding
        self._dims = self._set_model_dims()

    def _set_model_dims(self) -> int:
        try:
            embedding = self._model_client.create(
                input=["dimension test"],
                engine=self._model
            )["data"][0]["embedding"]
        except (KeyError, IndexError) as ke:
            raise ValueError(f"Unexpected response from the OpenAI API: {str(ke)}")
        except openai.error.AuthenticationError as ae:
            raise ValueError(f"Error authenticating with the OpenAI API: {str(ae)}")
        except Exception as e: # pylint: disable=broad-except
            # fall back (TODO get more specific)
            raise ValueError(f"Error setting embedding model dimensions: {str(e)}")
        return len(embedding)

    @retry(
        wait=wait_random_exponential(min=1, max=60),
        stop=stop_after_attempt(6),
        retry=retry_if_not_exception_type(TypeError),
    )
    def embed_many(
        self,
        texts: List[str],
        preprocess: Optional[Callable] = None,
        batch_size: Optional[int] = 10,
        as_buffer: Optional[float] = False,
    ) -> List[List[float]]:
        """Embed many chunks of texts using the OpenAI API.

        Args:
            texts (List[str]): List of text chunks to embed.
            preprocess (Optional[Callable], optional): Optional preprocessing callable to
                perform before vectorization. Defaults to None.
            batch_size (int, optional): Batch size of texts to use when creating
                embeddings. Defaults to 10.
            as_buffer (Optional[float], optional): Whether to convert the raw embedding
                to a byte string. Defaults to False.

        Returns:
            List[List[float]]: List of embeddings.

        Raises:
            TypeError: If the wrong input type is passed in for the test.
        """
        if not isinstance(texts, list):
                raise TypeError("Must pass in a list of str values to embed.")
        if  len(texts) > 0 and not isinstance(texts[0], str):
                raise TypeError("Must pass in a list of str values to embed.")

        embeddings: List = []
        for batch in self.batchify(texts, batch_size, preprocess):
            response = self._model_client.create(input=batch, engine=self._model)
            embeddings += [
                self._process_embedding(r["embedding"], as_buffer)
                for r in response["data"]
            ]
        return embeddings

    @retry(
        wait=wait_random_exponential(min=1, max=60),
        stop=stop_after_attempt(6),
        retry=retry_if_not_exception_type(TypeError),
    )
    def embed(
        self,
        text: str,
        preprocess: Optional[Callable] = None,
        as_buffer: Optional[float] = False,
    ) -> List[float]:
        """Embed a chunk of text using the OpenAI API.

        Args:
            text (str): Chunk of text to embed.
            preprocess (Optional[Callable], optional): Optional preprocessing callable to
                perform before vectorization. Defaults to None.
            as_buffer (Optional[float], optional): Whether to convert the raw embedding
                to a byte string. Defaults to False.

        Returns:
            List[float]: Embedding.

        Raises:
            TypeError: If the wrong input type is passed in for the test.
        """
        if not isinstance(text, str):
            raise TypeError("Must pass in a str value to embed.")

        if preprocess:
            text = preprocess(text)
        result = self._model_client.create(input=[text], engine=self._model)
        return self._process_embedding(result["data"][0]["embedding"], as_buffer)

    @retry(
        wait=wait_random_exponential(min=1, max=60),
        stop=stop_after_attempt(6),
        retry=retry_if_not_exception_type(TypeError),
    )
    async def aembed_many(
        self,
        texts: List[str],
        preprocess: Optional[Callable] = None,
        batch_size: int = 1000,
        as_buffer: Optional[bool] = False,
    ) -> List[List[float]]:
        """Asynchronously embed many chunks of texts using the OpenAI API.

        Args:
            texts (List[str]): List of text chunks to embed.
            preprocess (Optional[Callable], optional): Optional preprocessing callable to
                perform before vectorization. Defaults to None.
            batch_size (int, optional): Batch size of texts to use when creating
                embeddings. Defaults to 10.
            as_buffer (Optional[float], optional): Whether to convert the raw embedding
                to a byte string. Defaults to False.

        Returns:
            List[List[float]]: List of embeddings.

        Raises:
            TypeError: If the wrong input type is passed in for the test.
        """
        if not isinstance(texts, list):
                raise TypeError("Must pass in a list of str values to embed.")
        if  len(texts) > 0 and not isinstance(texts[0], str):
                raise TypeError("Must pass in a list of str values to embed.")

        embeddings: List = []
        for batch in self.batchify(texts, batch_size, preprocess):
            response = await self._model_client.acreate(input=batch, engine=self._model)
            embeddings += [
                self._process_embedding(r["embedding"], as_buffer)
                for r in response["data"]
            ]
        return embeddings

    @retry(
        wait=wait_random_exponential(min=1, max=60),
        stop=stop_after_attempt(6),
        retry=retry_if_not_exception_type(TypeError),
    )
    async def aembed(
        self,
        text: str,
        preprocess: Optional[Callable] = None,
        as_buffer: Optional[bool] = False,
    ) -> List[float]:
        """Asynchronously embed a chunk of text using the OpenAI API.

        Args:
            text (str): Chunk of text to embed.
            preprocess (Optional[Callable], optional): Optional preprocessing callable to
                perform before vectorization. Defaults to None.
            as_buffer (Optional[float], optional): Whether to convert the raw embedding
                to a byte string. Defaults to False.

        Returns:
            List[float]: Embedding.

        Raises:
            TypeError: If the wrong input type is passed in for the test.
        """
        if not isinstance(text, str):
            raise TypeError("Must pass in a str value to embed.")

        if preprocess:
            text = preprocess(text)
        result = await self._model_client.acreate(input=[text], engine=self._model)
        return self._process_embedding(result["data"][0]["embedding"], as_buffer)



from typing import Callable, List, Optional

from redisvl.utils.utils import array_to_buffer


class BaseVectorizer:
    _dims = None

    def __init__(self, model: str):
        self._model = model

    @property
    def model(self) -> str:
        return self._model

    @property
    def dims(self) -> int:
        return self._dims

    def set_model(self, model: str, dims: Optional[int] = None) -> None:
        self._model = model
        if dims is not None:
            self._dims = dims

    def embed_many(
        self,
        texts: List[str],
        preprocess: Optional[Callable] = None,
        batch_size: Optional[int] = 1000,
        as_buffer: Optional[bool] = False,
    ) -> List[List[float]]:
        raise NotImplementedError

    def embed(
        self,
        text: str,
        preprocess: Optional[Callable] = None,
        as_buffer: Optional[bool] = False,
    ) -> List[float]:
        raise NotImplementedError

    async def aembed_many(
        self,
        texts: List[str],
        preprocess: Optional[Callable] = None,
        batch_size: Optional[int] = 1000,
        as_buffer: Optional[bool] = False,
    ) -> List[List[float]]:
        raise NotImplementedError

    async def aembed(
        self,
        text: str,
        preprocess: Optional[Callable] = None,
        as_buffer: Optional[bool] = False,
    ) -> List[float]:
        raise NotImplementedError

    def batchify(self, seq: list, size: int, preprocess: Optional[Callable] = None):
        for pos in range(0, len(seq), size):
            if preprocess is not None:
                yield [preprocess(chunk) for chunk in seq[pos : pos + size]]
            else:
                yield seq[pos : pos + size]

    def _process_embedding(self, embedding: List[float], as_buffer: bool):
        if as_buffer:
            return array_to_buffer(embedding)
        return embedding



from typing import List, Optional, Union

from redis.commands.search.field import VectorField

from redisvl.index import SearchIndex
from redisvl.llmcache.base import BaseLLMCache
from redisvl.query import VectorQuery
from redisvl.utils.utils import array_to_buffer
from redisvl.vectorize.base import BaseVectorizer
from redisvl.vectorize.text import HFTextVectorizer


class SemanticCache(BaseLLMCache):
    """Cache for Large Language Models."""

    # TODO allow for user to change default fields
    _vector_field_name = "prompt_vector"
    _default_fields = [
        VectorField(
            _vector_field_name,
            "FLAT",
            {"DIM": 768, "TYPE": "FLOAT32", "DISTANCE_METRIC": "COSINE"},
        ),
    ]

    def __init__(
        self,
        index_name: Optional[str] = "cache",
        prefix: Optional[str] = "llmcache",
        threshold: Optional[float] = 0.9,
        ttl: Optional[int] = None,
        vectorizer: Optional[BaseVectorizer] = HFTextVectorizer(
            "sentence-transformers/all-mpnet-base-v2"
        ),
        redis_url: Optional[str] = "redis://localhost:6379",
        connection_args: Optional[dict] = None,
        index: Optional[SearchIndex] = None,
    ):
        """Semantic Cache for Large Language Models.

        Args:
            index_name (Optional[str], optional): The name of the index. Defaults to "cache".
            prefix (Optional[str], optional): The prefix for the index. Defaults to "llmcache".
            threshold (Optional[float], optional): Semantic threshold for the cache. Defaults to 0.9.
            ttl (Optional[int], optional): The TTL for the cache. Defaults to None.
            vectorizer (Optional[BaseVectorizer], optional): The vectorizer for the cache.
                Defaults to HFTextVectorizer("sentence-transformers/all-mpnet-base-v2").
            redis_url (Optional[str], optional): The redis url. Defaults to "redis://localhost:6379".
            connection_args (Optional[dict], optional): The connection arguments for the redis client. Defaults to None.
            index (Optional[SearchIndex], optional): The underlying search index to use for the semantic cache. Defaults to None.

        Raises:
            ValueError: If the threshold is not between 0 and 1.
            ValueError: If the index name or prefix is not supplied when constructing index manually.
        """
        self._ttl = ttl
        self._vectorizer = vectorizer
        self.set_threshold(threshold)
        connection_args = connection_args or {}

        if not index:
            if index_name and prefix:
                index = SearchIndex(
                    name=index_name, prefix=prefix, fields=self._default_fields
                )
                index.connect(url=redis_url, **connection_args)
            else:
                raise ValueError(
                    "Index name and prefix must be provided if not constructing from an existing index."
                )

        # create index if non-existent
        if not index.exists():
            index.create()

        self._index = index

    @classmethod
    def from_index(cls, index: SearchIndex, **kwargs):
        """Create a SemanticCache from a pre-existing SearchIndex.

        Args:
            index (SearchIndex): The SearchIndex object to use as the backbone of the cache.

        Returns:
            SemanticCache: A SemanticCache object.
        """
        return cls(index=index, **kwargs)

    @property
    def ttl(self) -> Optional[int]:
        """Returns the TTL for the cache.

        Returns:
            Optional[int]: The TTL for the cache.
        """
        return self._ttl

    def set_ttl(self, ttl: int):
        """Sets the TTL for the cache.

        Args:
            ttl (int): The TTL for the cache.

        Raises:
            ValueError: If the TTL is not an integer.
        """
        self._ttl = int(ttl)

    @property
    def index(self) -> SearchIndex:
        """Returns the index for the cache.

        Returns:
            SearchIndex: The index for the cache.
        """
        return self._index

    @property
    def threshold(self) -> float:
        """Returns the threshold for the cache."""
        return self._threshold

    def set_threshold(self, threshold: float):
        """Sets the threshold for the cache.

        Args:
            threshold (float): The threshold for the cache.

        Raises:
            ValueError: If the threshold is not between 0 and 1.
        """
        if not 0 <= float(threshold) <= 1:
            raise ValueError("Threshold must be between 0 and 1.")
        self._threshold = float(threshold)

    def clear(self):
        """Clear the LLMCache of all keys in the index"""
        client = self._index.client
        if client:
            with client.pipeline(transaction=False) as pipe:
                for key in client.scan_iter(match=f"{self._index._prefix}:*"):
                    pipe.delete(key)
                pipe.execute()
        else:
            raise RuntimeError("LLMCache is not connected to a Redis instance.")

    def check(
        self,
        prompt: Optional[str] = None,
        vector: Optional[List[float]] = None,
        num_results: int = 1,
        fields: List[str] = ["response"],
    ) -> List[str]:
        """Checks whether the cache contains the specified prompt or vector.

        Args:
            prompt (Optional[str], optional): The prompt to check. Defaults to None.
            vector (Optional[List[float]], optional): The vector to check. Defaults to None.
            num_results (int, optional): The number of results to return. Defaults to 1.
            fields (List[str], optional): The fields to return. Defaults to ["response"].

        Raises:
            ValueError: If neither prompt nor vector is specified.

        Returns:
            List[str]: The response(s) if the cache contains the prompt or vector.
        """
        if not prompt and not vector:
            raise ValueError("Either prompt or vector must be specified.")

        if not vector:
            vector = self._vectorizer.embed(prompt)  # type: ignore

        v = VectorQuery(
            vector=vector,
            vector_field_name=self._vector_field_name,
            return_fields=fields,
            num_results=num_results,
            return_score=True,
        )

        cache_hits: List[str] = []
        results = self._index.search(v.query, query_params=v.params)
        for result in results.docs:
            sim = similarity(result["vector_distance"])
            if sim > self.threshold:
                self._refresh_ttl(result["id"])
                cache_hits.append(
                    result["response"]
                )  # TODO - in the future what do we actually want to return here?
        return cache_hits

    def store(
        self,
        prompt: str,
        response: str,
        vector: Optional[List[float]] = None,
        metadata: Optional[dict] = {},
    ) -> None:
        """Stores the specified key-value pair in the cache along with metadata.

        Args:
            prompt (str): The prompt to store.
            response (str): The response to store.
            vector (Optional[List[float]], optional): The vector to store. Defaults to None.
            metadata (Optional[dict], optional): The metadata to store. Defaults to {}.

        Raises:
            ValueError: If neither prompt nor vector is specified.
        """
        # TODO - foot gun for schema mismatch if user has a different index
        vector = vector or self._vectorizer.embed(prompt)
        payload = {
            "id": self.hash_input(prompt),
            "prompt": prompt,
            "response": response,
            self._vector_field_name: array_to_buffer(vector),
        }
        if metadata:
            payload.update(metadata)

        # Load LLMCache entry with TTL
        self._index.load([payload], ttl=self._ttl, key_field="id")

    def _refresh_ttl(self, key: str):
        """Refreshes the TTL for the specified key."""
        client = self._index.client
        if client:
            if self.ttl:
                client.expire(key, self.ttl)
        else:
            raise RuntimeError("LLMCache is not connected to a Redis instance.")


def similarity(distance: Union[float, str]) -> float:
    return 1 - float(distance)



import hashlib
from typing import Callable, List, Optional

from redisvl.index import SearchIndex


class BaseLLMCache:
    verbose: bool = True

    @classmethod
    def from_index(cls, index: SearchIndex, **kwargs):
        """Create a SemanticCache from a pre-existing SearchIndex."""
        raise NotImplementedError

    def clear(self):
        """Clear the LLMCache and create a new underlying index."""
        raise NotImplementedError

    def check(self, prompt: str) -> Optional[List[str]]:
        raise NotImplementedError

    def store(
        self,
        prompt: str,
        response: str,
        vector: Optional[List[float]] = None,
        metadata: Optional[dict] = {},
    ) -> None:
        """Stores the specified key-value pair in the cache along with metadata."""
        raise NotImplementedError

    def _refresh_ttl(self, key: str):
        """Refreshes the TTL for the specified key."""
        raise NotImplementedError

    def hash_input(self, prompt: str):
        """Hashes the input using SHA256."""
        return hashlib.sha256(prompt.encode("utf-8")).hexdigest()



from typing import TYPE_CHECKING, Any, Dict, List, Optional

import numpy as np
from redis.commands.search.query import Query

from redisvl.query.filter import FilterExpression
from redisvl.utils.utils import array_to_buffer


class BaseQuery:
    def __init__(
        self, return_fields: Optional[List[str]] = None, num_results: Optional[int] = 10
    ):
        self._return_fields = return_fields
        self._num_results = num_results

    @property
    def query(self) -> "Query":
        pass

    @property
    def params(self) -> Dict[str, Any]:
        pass


class VectorQuery(BaseQuery):
    dtypes = {
        "float32": np.float32,
        "float64": np.float64,
    }

    DISTANCE_ID = "vector_distance"

    def __init__(
        self,
        vector: List[float],
        vector_field_name: str,
        return_fields: List[str],
        filter_expression: FilterExpression = None,
        dtype: str = "float32",
        num_results: Optional[int] = 10,
        return_score: bool = True,
    ):
        """Query for vector fields

        Args:
            vector (List[float]): The vector to query for.
            vector_field_name (str): The name of the vector field
            return_fields (List[str]): The fields to return.
            filter_expression (FilterExpression, optional): A filter to apply to the query. Defaults to None.
            dtype (str, optional): The dtype of the vector. Defaults to "float32".
            num_results (Optional[int], optional): The number of results to return. Defaults to 10.
            return_score (bool, optional): Whether to return the score. Defaults to True.

        Raises:
            TypeError: If filter_expression is not of type redisvl.query.FilterExpression

        """
        super().__init__(return_fields, num_results)
        self._vector = vector
        self._field = vector_field_name
        self._dtype = dtype.lower()
        if filter_expression:
            self.set_filter(filter_expression)
        else:
            self._filter = "*"

        if return_score:
            self._return_fields.append(self.DISTANCE_ID)

    def set_filter(self, filter_expression: FilterExpression):
        """Set the filter for the query.

        Args:
            filter_expression (FilterExpression): The filter to apply to the query.
        """
        if not isinstance(filter_expression, FilterExpression):
            raise TypeError(
                "filter_expression must be of type redisvl.query.FilterExpression"
            )
        self._filter = str(filter_expression)

    def get_filter(self) -> FilterExpression:
        """Get the filter for the query.

        Returns:
            FilterExpression: The filter for the query.
        """
        return self._filter

    def __str__(self):
        return " ".join([str(x) for x in self.query.get_args()])

    @property
    def query(self) -> Query:
        """Return a Redis-Py Query object representing the query.

        Returns:
            redis.commands.search.query.Query: The query object.
        """
        base_query = f"{self._filter}=>[KNN {self._num_results} @{self._field} $vector AS {self.DISTANCE_ID}]"
        query = (
            Query(base_query)
            .return_fields(*self._return_fields)
            .sort_by(self.DISTANCE_ID)
            .paging(0, self._num_results)
            .dialect(2)
        )
        return query

    @property
    def params(self) -> Dict[str, Any]:
        """Return the parameters for the query.

        Returns:
            Dict[str, Any]: The parameters for the query.
        """
        return {"vector": array_to_buffer(self._vector, dtype=self.dtypes[self._dtype])}



import logging
import sys

import coloredlogs

# constants for logging
coloredlogs.DEFAULT_DATE_FORMAT = "%H:%M:%S"
coloredlogs.DEFAULT_LOG_FORMAT = "%(asctime)s %(name)s %(levelname)s   %(message)s"


def get_logger(name, log_level="info", fmt=None):
    """Return a logger instance"""

    # Use file name if logger is in debug mode
    name = "RedisVL" if log_level == "debug" else name

    logger = logging.getLogger(name)
    coloredlogs.install(level=log_level, logger=logger, fmt=fmt, stream=sys.stdout)
    return logger



import os
from functools import wraps
from typing import Optional

# TODO: handle connection errors.


def get_redis_connection(url: Optional[str] = None, **kwargs):
    from redis import Redis

    if url:
        client = Redis.from_url(url, **kwargs)
    else:
        try:
            client = Redis.from_url(get_address_from_env())
        except ValueError:
            raise ValueError("No Redis URL provided and REDIS_URL env var not set")
    return client


def get_async_redis_connection(url: Optional[str] = None, **kwargs):
    from redis.asyncio import Redis as ARedis

    if url:
        client = ARedis.from_url(url, **kwargs)
    else:
        try:
            client = ARedis.from_url(get_address_from_env())
        except ValueError:
            raise ValueError("No Redis URL provided and REDIS_URL env var not set")
    return client


def get_address_from_env():
    """Get a redis connection from environment variables

    Returns:
        str: Redis URL
    """
    addr = os.getenv("REDIS_URL", None)
    if not addr:
        raise ValueError("REDIS_URL env var not set")
    return addr


def check_connected(client_variable_name: str):
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if getattr(self, client_variable_name) is None:
                raise ValueError(
                    f"SearchIndex.connect() must be called before calling {func.__name__}"
                )
            return func(self, *args, **kwargs)

        return wrapper

    return decorator



import argparse
import sys
from tabulate import tabulate
from argparse import Namespace

from redisvl.cli.utils import create_redis_url, add_index_parsing_options
from redisvl.index import SearchIndex
from redisvl.utils.connection import get_redis_connection

from redisvl.cli.log import get_logger
logger = get_logger("[RedisVL]")

STATS_KEYS = [
    "num_docs",
    "num_terms",
    "max_doc_id",
    "num_records",
    "percent_indexed",
    "hash_indexing_failures",
    "number_of_uses",
    "bytes_per_record_avg",
    "doc_table_size_mb",
    "inverted_sz_mb",
    "key_table_size_mb",
    "offset_bits_per_record_avg",
    "offset_vectors_sz_mb",
    "offsets_per_term_avg",
    "records_per_doc_avg",
    "sortable_values_size_mb",
    "total_indexing_time",
    "total_inverted_index_blocks",
    "vector_index_sz_mb",
]

class Stats:
    usage = "\n".join(
        [
            "rvl stats [<args>]\n",
        ]
    )

    def __init__(self):
        parser = argparse.ArgumentParser(usage=self.usage)

        parser.add_argument(
            "-f",
            "--format",
            help="Output format",
            type=str,
            default="rounded_outline"
        )
        parser = add_index_parsing_options(parser)
        args = parser.parse_args(sys.argv[2:])
        try:
            self.stats(args)
        except Exception as e:
            logger.error(e)
            exit(0)


    def stats(self, args: Namespace):
        """Obtain stats about an index

        Usage:
            rvl stats -i <index_name> | -s <schema_path>
        """
        index = self._connect_to_index(args)
        _display_stats(index.info(), output_format=args.format)

    def _connect_to_index(self, args: Namespace) -> SearchIndex:
        # connect to redis
        try:
            url = create_redis_url(args)
            conn = get_redis_connection(url=url)
        except ValueError:
            logger.error(
                "Must set REDIS_ADDRESS environment variable or provide host and port"
            )
            exit(0)

        if args.index:
            index = SearchIndex.from_existing(name=args.index, url=url)
        elif args.schema:
            index = SearchIndex.from_yaml(args.schema)
            index.set_client(conn)
        else:
            logger.error("Index name or schema must be provided")
            exit(0)

        return index


def _display_stats(index_info, output_format="rounded_outline"):
    # Extracting the statistics
    stats_data = [(key, str(index_info.get(key))) for key in STATS_KEYS]

    # Display the statistics in tabular format
    print("\nStatistics:")
    print(
        tabulate(
            stats_data,
            headers=["Stat Key", "Value"],
            tablefmt=output_format,
            colalign=("left", "left"),
        )
    )



from pathlib import Path
from typing import List, Optional, Union
from uuid import uuid4

import yaml
from pydantic import BaseModel, Field, field_validator
from redis.commands.search.field import (
    GeoField,
    NumericField,
    TagField,
    TextField,
    VectorField,
)
from typing_extensions import Literal


class BaseField(BaseModel):
    name: str = Field(...)
    sortable: Optional[bool] = False


class TextFieldSchema(BaseField):
    weight: Optional[float] = 1
    no_stem: Optional[bool] = False
    phonetic_matcher: Optional[str] = None
    withsuffixtrie: Optional[bool] = False

    def as_field(self):
        return TextField(
            self.name,
            weight=self.weight,
            no_stem=self.no_stem,
            phonetic_matcher=self.phonetic_matcher,
            sortable=self.sortable,
        )


class TagFieldSchema(BaseField):
    separator: Optional[str] = ","
    case_sensitive: Optional[bool] = False

    def as_field(self):
        return TagField(
            self.name,
            separator=self.separator,
            case_sensitive=self.case_sensitive,
            sortable=self.sortable,
        )


class NumericFieldSchema(BaseField):
    def as_field(self):
        return NumericField(self.name, sortable=self.sortable)


class GeoFieldSchema(BaseField):
    def as_field(self):
        return GeoField(self.name, sortable=self.sortable)


class BaseVectorField(BaseModel):
    name: str = Field(...)
    dims: int = Field(...)
    algorithm: object = Field(...)
    datatype: str = Field(default="FLOAT32")
    distance_metric: str = Field(default="COSINE")
    initial_cap: int = Field(default=20000)

    @field_validator("algorithm", "datatype", "distance_metric", mode="before")
    @classmethod
    def uppercase_strings(cls, v):
        return v.upper()


class FlatVectorField(BaseVectorField):
    algorithm: object = Literal["FLAT"]
    block_size: int = Field(default=1000)

    def as_field(self):
        return VectorField(
            self.name,
            self.algorithm,
            {
                "TYPE": self.datatype,
                "DIM": self.dims,
                "DISTANCE_METRIC": self.distance_metric,
                "INITIAL_CAP": self.initial_cap,
                "BLOCK_SIZE": self.block_size,
            },
        )


class HNSWVectorField(BaseVectorField):
    algorithm: object = Literal["HNSW"]
    m: int = Field(default=16)
    ef_construction: int = Field(default=200)
    ef_runtime: int = Field(default=10)
    epsilon: float = Field(default=0.8)

    def as_field(self):
        return VectorField(
            self.name,
            self.algorithm,
            {
                "TYPE": self.datatype,
                "DIM": self.dims,
                "DISTANCE_METRIC": self.distance_metric,
                "INITIAL_CAP": self.initial_cap,
                "M": self.m,
                "EF_CONSTRUCTION": self.ef_construction,
                "EF_RUNTIME": self.ef_runtime,
                "EPSILON": self.epsilon,
            },
        )


class IndexModel(BaseModel):
    name: str = Field(...)
    prefix: str = Field(...)
    storage_type: Optional[str] = Field(default="hash")


class FieldsModel(BaseModel):
    tag: Optional[List[TagFieldSchema]] = None
    text: Optional[List[TextFieldSchema]] = None
    numeric: Optional[List[NumericFieldSchema]] = None
    geo: Optional[List[GeoFieldSchema]] = None
    vector: Optional[List[Union[FlatVectorField, HNSWVectorField]]] = None


class SchemaModel(BaseModel):
    index: IndexModel = Field(...)
    fields: FieldsModel = Field(...)

    @field_validator("index")
    @classmethod
    def validate_index(cls, v):
        if v.storage_type not in ["hash", "json"]:
            raise ValueError(f"Storage type {v.storage_type} not supported")
        return v

    @property
    def index_fields(self):
        redis_fields = []
        for field_name in self.fields.model_fields.keys():
            field_group = getattr(self.fields, field_name)
            if field_group is not None:
                for field in field_group:
                    redis_fields.append(field.as_field())
        return redis_fields


def read_schema(file_path: str):
    fp = Path(file_path).resolve()
    if not fp.exists():
        raise FileNotFoundError(f"Schema file {file_path} does not exist")

    with open(fp, "r") as f:
        schema = yaml.safe_load(f)

    return SchemaModel(**schema)



import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Pattern

if TYPE_CHECKING:
    from redis.commands.search.result import Result
    from redis.commands.search.document import Document

import numpy as np


class TokenEscaper:
    """
    Escape punctuation within an input string. Taken from RedisOM Python.
    """

    # Characters that RediSearch requires us to escape during queries.
    # Source: https://redis.io/docs/stack/search/reference/escaping/#the-rules-of-text-field-tokenization
    DEFAULT_ESCAPED_CHARS = r"[,.<>{}\[\]\\\"\':;!@#$%^&*()\-+=~\/ ]"

    def __init__(self, escape_chars_re: Optional[Pattern] = None):
        if escape_chars_re:
            self.escaped_chars_re = escape_chars_re
        else:
            self.escaped_chars_re = re.compile(self.DEFAULT_ESCAPED_CHARS)

    def escape(self, value: str) -> str:
        def escape_symbol(match):
            value = match.group(0)
            return f"\\{value}"

        return self.escaped_chars_re.sub(escape_symbol, value)


def make_dict(values: List[Any]):
    # TODO make this a real function
    i = 0
    di = {}
    while i < len(values) - 1:
        di[values[i]] = values[i + 1]
        i += 2
    return di


def convert_bytes(data: Any) -> Any:
    if isinstance(data, bytes):
        return data.decode("ascii")
    if isinstance(data, dict):
        return dict(map(convert_bytes, data.items()))
    if isinstance(data, list):
        return list(map(convert_bytes, data))
    if isinstance(data, tuple):
        return map(convert_bytes, data)
    return data


# required modules
REDIS_REQUIRED_MODULES = [
    {"name": "search", "ver": 20400},
    {"name": "searchlight", "ver": 20400},
]


def check_redis_modules_exist(client) -> None:
    """Check if the correct Redis modules are installed."""
    installed_modules = client.module_list()
    installed_modules = {
        module[b"name"].decode("utf-8"): module for module in installed_modules
    }
    for module in REDIS_REQUIRED_MODULES:
        if module["name"] in installed_modules and int(
            installed_modules[module["name"]][b"ver"]
        ) >= int(
            module["ver"]
        ):  # type: ignore[call-overload]
            return
    # otherwise raise error
    error_message = (
        "You must add the RediSearch (>= 2.4) module from Redis Stack. "
        "Please refer to Redis Stack docs: https://redis.io/docs/stack/"
    )
    raise ValueError(error_message)


def array_to_buffer(array: List[float], dtype: Any = np.float32) -> bytes:
    """Convert a list of floats into a numpy byte string."""
    return np.array(array).astype(dtype).tobytes()


def process_results(results: List["Result"]) -> List[Dict[str, Any]]:
    """Convert a list of search Result objects into a list of document dicts"""

    def _process(doc: "Document") -> dict:
        d = doc.__dict__
        if "payload" in d:
            del d["payload"]
        return d

    return [_process(doc) for doc in results.docs]



from enum import Enum
from typing import Any, List, Optional, Union

from redisvl.utils.utils import TokenEscaper


class FilterOperator(Enum):
    EQ = 1
    NE = 2
    LT = 3
    GT = 4
    LE = 5
    GE = 6
    OR = 7
    AND = 8
    LIKE = 9
    IN = 10


class FilterField:
    escaper = TokenEscaper()
    OPERATORS = []

    def __init__(self, field: str):
        self._field = field
        self._value = None
        self._operator = None

    def _set_value(self, val: Any, val_type: type, operator: FilterOperator):
        # check that the operator is supported by this class
        if operator not in self.OPERATORS:
            raise ValueError(
                f"Operator {operator} not supported by {self.__class__.__name__}. "
                + f"Supported operators are {self.OPERATORS.values()}"
            )

        if not isinstance(val, val_type):
            raise TypeError(
                f"Right side argument passed to operator {self.OPERATORS[operator]} with left side "
                f"argument {self.__class__.__name__} must be of type {val_type}"
            )
        self._value = val
        self._operator = operator


class Tag(FilterField):
    """A Tag is a FilterField representing a tag in a Redis index."""

    OPERATORS = {
        FilterOperator.EQ: "==",
        FilterOperator.NE: "!=",
        FilterOperator.IN: "==",
    }

    OPERATOR_MAP = {
        FilterOperator.EQ: "@%s:{%s}",
        FilterOperator.NE: "(-@%s:{%s})",
        FilterOperator.IN: "@%s:{%s}",
    }

    def __init__(self, field: str):
        """Create a Tag FilterField

        Args:
            field (str): The name of the tag field in the index to be queried against
        """
        super().__init__(field)

    def _set_tag_value(self, other: Union[List[str], str], operator: FilterOperator):
        if isinstance(other, list):
            if not all(isinstance(tag, str) for tag in other):
                raise ValueError("All tags must be strings")
        else:
            other = [other]
        self._set_value(other, list, operator)

    def __eq__(self, other: Union[List[str], str]) -> "FilterExpression":
        """Create a Tag equality filter expression

        Args:
            other (Union[List[str], str]): The tag(s) to filter on.

        Example:
            >>> from redisvl.query.filter import Tag
            >>> filter = Tag("brand") == "nike"
        """
        self._set_tag_value(other, FilterOperator.EQ)
        return FilterExpression(str(self))

    def __ne__(self, other) -> "FilterExpression":
        """Create a Tag inequality filter expression

        Args:
            other (Union[List[str], str]): The tag(s) to filter on.

        Example:
            >>> from redisvl.query.filter import Tag
            >>> filter = Tag("brand") != "nike"
        """
        self._set_tag_value(other, FilterOperator.NE)
        return FilterExpression(str(self))

    @property
    def _formatted_tag_value(self) -> str:
        return "|".join([self.escaper.escape(tag) for tag in self._value])

    def __str__(self) -> str:
        """Return the Redis Query syntax for a Tag filter expression"""
        return self.OPERATOR_MAP[self._operator] % (
            self._field,
            self._formatted_tag_value,
        )


class Geo(FilterField):
    """A Geo is a FilterField representing a geographic (lat/lon)
    field in a Redis index.

    """
    OPERATORS = {
        FilterOperator.EQ: "==",
        FilterOperator.NE: "!=",
    }
    OPERATOR_MAP = {
        FilterOperator.EQ: "@%s:[%f %f %i %s]",
        FilterOperator.NE: "(-@%s:[%f %f %i %s])",
    }

    def __eq__(self, other) -> "FilterExpression":
        """Create a Geographic equality filter expression

        Args:
            other (GeoSpec): The geographic spec to filter on.

        Example:
            >>> from redisvl.query.filter import Geo, GeoRadius
            >>> filter = Geo("location") == GeoRadius(-122.4194, 37.7749, 1, unit="m")
        """
        self._set_value(other, GeoSpec, FilterOperator.EQ)
        return FilterExpression(str(self))

    def __ne__(self, other) -> "FilterExpression":
        """Create a Geographic inequality filter expression

        Args:
            other (GeoSpec): The geographic spec to filter on.

        Example:
            >>> from redisvl.query.filter import Geo, GeoRadius
            >>> filter = Geo("location") != GeoRadius(-122.4194, 37.7749, 1, unit="m")
        """
        self._set_value(other, GeoSpec, FilterOperator.NE)
        return FilterExpression(str(self))

    def __str__(self) -> "FilterExpression":
        """Return the Redis Query syntax for a Geographic filter expression"""
        return self.OPERATOR_MAP[self._operator] % (
            self._field,
            *self._value.get_args(),
        )


class GeoSpec:
    GEO_UNITS = ["m", "km", "mi", "ft"]

    # class for the operand for FilterExpressions with Geo
    def __init__(self, longitude: float, latitude: float, unit: str = "km"):
        if unit.lower() not in self.GEO_UNITS:
            raise ValueError(f"Unit must be one of {self.GEO_UNITS}")
        self._longitude = longitude
        self._latitude = latitude
        self._unit = unit.lower()


class GeoRadius(GeoSpec):
    """A GeoRadius is a GeoSpec representing a geographic radius"""
    def __init__(
        self,
        longitude: float,
        latitude: float,
        radius: Optional[int] = 1,
        unit: Optional[str] = "km"
    ):
        """Create a GeoRadius specification (GeoSpec)

        Args:
            longitude (float): The longitude of the center of the radius.
            latitude (float): The latitude of the center of the radius.
            radius (int, optional): The radius of the circle. Defaults to 1.
            unit (str, optional): The unit of the radius. Defaults to "km".

        Raises:
            ValueError: If the unit is not one of "m", "km", "mi", or "ft".

        """
        super().__init__(longitude, latitude, unit)
        self._radius = radius

    def get_args(self) -> List[Union[float, int, str]]:
        return [self._longitude, self._latitude, self._radius, self._unit]


class Num(FilterField):
    """A Num is a FilterField representing a numeric field in a Redis index."""
    OPERATORS = {
        FilterOperator.EQ: "==",
        FilterOperator.NE: "!=",
        FilterOperator.LT: "<",
        FilterOperator.GT: ">",
        FilterOperator.LE: "<=",
        FilterOperator.GE: ">=",
    }
    OPERATOR_MAP = {
        FilterOperator.EQ: "@%s:[%i %i]",
        FilterOperator.NE: "(-@%s:[%i %i])",
        FilterOperator.GT: "@%s:[(%i +inf]",
        FilterOperator.LT: "@%s:[-inf (%i]",
        FilterOperator.GE: "@%s:[%i +inf]",
        FilterOperator.LE: "@%s:[-inf %i]",
    }

    def __str__(self) -> str:
        """Return the Redis Query syntax for a Numeric filter expression"""
        if self._operator == FilterOperator.EQ or self._operator == FilterOperator.NE:
            return self.OPERATOR_MAP[self._operator] % (
                self._field,
                self._value,
                self._value,
            )
        else:
            return self.OPERATOR_MAP[self._operator] % (self._field, self._value)

    def __eq__(self, other: str) -> "FilterExpression":
        """Create a Numeric equality filter expression

        Args:
            other (int): The value to filter on.

        Example:
            >>> from redisvl.query.filter import Num
            >>> filter = Num("zipcode") == 90210
        """
        self._set_value(other, int, FilterOperator.EQ)
        return FilterExpression(str(self))

    def __ne__(self, other: str) -> "FilterExpression":
        """Create a Numeric inequality filter expression

        Args:
            other (int): The value to filter on.

        Example:
            >>> from redisvl.query.filter import Num
            >>> filter = Num("zipcode") != 90210
        """
        self._set_value(other, int, FilterOperator.NE)
        return FilterExpression(str(self))

    def __gt__(self, other: str) -> "FilterExpression":
        """Create a Numeric greater than filter expression

        Args:
            other (int): The value to filter on.

        Example:
            >>> from redisvl.query.filter import Num
            >>> filter = Num("age") > 18
        """
        self._set_value(other, int, FilterOperator.GT)
        return FilterExpression(str(self))

    def __lt__(self, other: str) -> "FilterExpression":
        """Create a Numeric less than filter expression

        Args:
            other (int): The value to filter on.

        Example:
            >>> from redisvl.query.filter import Num
            >>> filter = Num("age") < 18
        """
        self._set_value(other, int, FilterOperator.LT)
        return FilterExpression(str(self))

    def __ge__(self, other: str) -> "FilterExpression":
        """Create a Numeric greater than or equal to filter expression

        Args:
            other (int): The value to filter on.

        Example:
            >>> from redisvl.query.filter import Num
            >>> filter = Num("age") >= 18
        """
        self._set_value(other, int, FilterOperator.GE)
        return FilterExpression(str(self))

    def __le__(self, other: str) -> "FilterExpression":
        """Create a Numeric less than or equal to filter expression

        Args:
            other (int): The value to filter on.

        Example:
            >>> from redisvl.query.filter import Num
            >>> filter = Num("age") <= 18
        """
        self._set_value(other, int, FilterOperator.LE)
        return FilterExpression(str(self))


class Text(FilterField):
    """A Text is a FilterField representing a text field in a Redis index."""
    OPERATORS = {
        FilterOperator.EQ: "==",
        FilterOperator.NE: "!=",
        FilterOperator.LIKE: "%",
    }
    OPERATOR_MAP = {
        FilterOperator.EQ: '@%s:"%s"',
        FilterOperator.NE: '(-@%s:"%s")',
        FilterOperator.LIKE: "@%s:%s",
    }

    def __eq__(self, other: str) -> "FilterExpression":
        """Create a Text equality filter expression

        Args:
            other (str): The text value to filter on.

        Example:
            >>> from redisvl.query.filter import Text
            >>> filter = Text("job") == "engineer"
        """
        self._set_value(other, str, FilterOperator.EQ)
        return FilterExpression(str(self))

    def __ne__(self, other: str) -> "FilterExpression":
        """Create a Text inequality filter expression

        Args:
            other (str): The text value to filter on.

        Example:
            >>> from redisvl.query.filter import Text
            >>> filter = Text("job") != "engineer"
        """
        self._set_value(other, str, FilterOperator.NE)
        return FilterExpression(str(self))

    def __mod__(self, other: str) -> "FilterExpression":
        """Create a Text like filter expression

        Args:
            other (str): The text value to filter on.

        Example:
            >>> from redisvl.query.filter import Text
            >>> filter = Text("job") % "engineer"
        """
        self._set_value(other, str, FilterOperator.LIKE)
        return FilterExpression(str(self))

    def __str__(self) -> str:
        try:
            return self.OPERATOR_MAP[self._operator] % (self._field, self._value)
        except KeyError:
            raise Exception("Invalid operator")


class FilterExpression:
    """A FilterExpression is a logical expression of FilterFields.

    FilterExpressions can be combined using the & and | operators to create
    complex logical expressions that evaluate to the Redis Query language.

    This presents an interface by which users can create complex queries
    without having to know the Redis Query language.

    Filter expressions are not created directly. Instead they are built
    by combining FilterFields using the & and | operators.

    Examples:

        >>> from redisvl.query.filter import Tag, Num
        >>> brand_is_nike = Tag("brand") == "nike"
        >>> price_is_over_100 = Num("price") < 100
        >>> filter = brand_is_nike & price_is_over_100
        >>> print(str(filter))
        (@brand:{nike} @price:[-inf (100)])

    This can be combined with the VectorQuery class to create a query:

        >>> from redisvl.query import VectorQuery
        >>> v = VectorQuery(
        ...     vector=[0.1, 0.1, 0.5, ...],
        ...     vector_field_name="product_embedding",
        ...     return_fields=["product_id", "brand", "price"],
        ...     filter_expression=filter,
        ... )
    """
    def __init__(
        self,
        _filter: str = None,
        operator: FilterOperator = None,
        left: Optional["FilterExpression"] = None,
        right: Optional["FilterExpression"] = None,
    ):
        self._filter = _filter
        self._operator = operator
        self._left = left
        self._right = right

    def __and__(self, other):
        return FilterExpression(operator=FilterOperator.AND, left=self, right=other)

    def __or__(self, other):
        return FilterExpression(operator=FilterOperator.OR, left=self, right=other)

    def __str__(self):
        if self._operator:
            operator_str = " | " if self._operator == FilterOperator.OR else " "
            return f"({str(self._left)}{operator_str}{str(self._right)})"
        return self._filter



from setuptools import setup, find_packages

# function to read in requirements.txt to into a python list
def read_requirements():
    with open("requirements.txt") as f:
        requirements = f.read().splitlines()
    return requirements

def read_dev_requirements():
    with open("requirements-dev.txt") as f:
        requirements = f.read().splitlines()
    return requirements

extras_require = {
    "all": [
        "openai>=0.26.4",
        "sentence-transformers>=2.2.2"
    ],
    "dev": read_dev_requirements()
}


setup(
    name="redisvl",
    version="0.0.3",
    python_requires=">=3.7",
    install_requires=read_requirements(),
    extras_require=extras_require,
    packages=find_packages(),
    package_data = {"redisvl": ["requirements.txt", "requirements-dev.txt"]},
    zip_safe=False,
    entry_points={
        "console_scripts": [
            "rvl = redisvl.cli.__init__:main"
        ]
    }
)


import os
from argparse import Namespace, ArgumentParser


def create_redis_url(args: Namespace) -> str:
    env_address = os.getenv("REDIS_URL")
    if env_address:
        return env_address
    else:
        url = "redis://"
        if args.ssl:
            url += "rediss://"
        if args.user:
            url += args.user
            if args.password:
                url += ":" + args.password
            url += "@"
        url += args.host + ":" + str(args.port)
        return url

def add_index_parsing_options(parser: ArgumentParser) -> ArgumentParser:
    parser.add_argument(
        "-i", "--index", help="Index name", type=str, required=False
    )
    parser.add_argument(
        "-s", "--schema", help="Path to schema file", type=str, required=False
    )
    parser.add_argument("--host", help="Redis host", type=str, default="localhost")
    parser.add_argument("-p", "--port", help="Redis port", type=int, default=6379)
    parser.add_argument(
        "--user", help="Redis username", type=str, default="default"
    )
    parser.add_argument("--ssl", help="Use SSL", action="store_true")
    parser.add_argument(
        "-a", "--password", help="Redis password", type=str, default=""
    )
    return parser



import asyncio
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional
from uuid import uuid4

if TYPE_CHECKING:
    from redis.commands.search.field import Field
    from redis.commands.search.result import Result
    from redisvl.query.query import BaseQuery

import redis
from redis.commands.search.indexDefinition import IndexDefinition, IndexType

from redisvl.schema import SchemaModel, read_schema
from redisvl.utils.connection import (
    check_connected,
    get_async_redis_connection,
    get_redis_connection,
)
from redisvl.utils.utils import (
    check_redis_modules_exist,
    convert_bytes,
    make_dict,
    process_results,
)


class SearchIndexBase:
    def __init__(
        self,
        name: str,
        prefix: str = "rvl",
        storage_type: Optional[str] = "hash",
        fields: Optional[List["Field"]] = None,
    ):
        self._name = name
        self._prefix = prefix
        self._storage = storage_type
        self._fields = fields
        self._redis_conn: Optional[redis.Redis] = None

    def set_client(self, client: redis.Redis):
        self._redis_conn = client

    @property
    @check_connected("_redis_conn")
    def client(self) -> redis.Redis:
        """The redis-py client object.

        Returns:
            redis.Redis: The redis-py client object
        """
        return self._redis_conn  # type: ignore

    @check_connected("_redis_conn")
    def search(self, *args, **kwargs) -> List["Result"]:
        """Perform a search on this index.

        Wrapper around redis.search.Search that adds the index name
        to the search query and passes along the rest of the arguments
        to the redis-py ft.search() method.

        Returns:
            List[Result]: A list of search results
        """
        results: List["Result"] = self._redis_conn.ft(self._name).search(
            *args, **kwargs
        )
        return results

    @check_connected("_redis_conn")
    def query(self, query: "BaseQuery") -> List[Dict[str, Any]]:
        """Run a query on this index.

        This is similar to the search method, but takes a BaseQuery
        object directly (does not allow for the usage of a raw
        redis query string) and post-processes results of the search.

        Args:
            query (BaseQuery): The query to run.

        Returns:
            List[Result]: A list of search results.
        """
        results = self.search(query.query, query_params=query.params)
        return process_results(results)

    @classmethod
    def from_yaml(cls, schema_path: str):
        """Create a SearchIndex from a YAML schema file.

        Args:
            schema_path (str): Path to the YAML schema file.

        Returns:
            SearchIndex: A SearchIndex object.
        """
        schema = read_schema(schema_path)
        return cls(fields=schema.index_fields, **schema.index.model_dump())

    @classmethod
    def from_dict(cls, schema_dict: Dict[str, Any]):
        """Create a SearchIndex from a dictionary.

        Args:
            schema_dict (Dict[str, Any]): A dictionary containing the schema.

        Returns:
            SearchIndex: A SearchIndex object.
        """
        schema = SchemaModel(**schema_dict)
        return cls(fields=schema.index_fields, **schema.index.model_dump())

    @classmethod
    def from_existing(
        cls,
        name: str,
        url: Optional[str] = None,
        fields: Optional[List["Field"]] = None,
        **kwargs,
    ):
        """Create a SearchIndex from an existing index in Redis.

        Args:
            name (str): Index name.
            url (Optional[str], optional): Redis URL. REDIS_URL env var
                is used if not provided. Defaults to None.
            fields (Optional[List[Field]], optional): List of Redis search
                fields to include in the schema. Defaults to None.

        Returns:
            SearchIndex: A SearchIndex object.

        Raises:
            redis.exceptions.ResponseError: If the index does not exist.
            ValueError: If the REDIS_URL env var is not set and url is not provided.
        """
        raise NotImplementedError

    def connect(self, url: str, **kwargs):
        """Connect to a Redis instance.

        Args:
            url (str): Redis URL. REDIS_URL env var is used if not provided.
        """
        raise NotImplementedError

    def disconnect(self):
        """Disconnect from the Redis instance"""
        self._redis_conn = None

    def _get_key(self, record: Dict[str, Any], key_field: str = None) -> str:
        """Construct the Redis HASH top level key.

        Args:
            record (Dict[str, Any]): A dictionary containing the record to be indexed.
            key_field (Optional[str], optional): A field within the record
                to use in the Redis hash key.

        Returns:
            str: The key to be used for a given record in Redis.

        Raises:
            ValueError: If the key field is not found in the record.
        """
        if key_field is None:
            key = uuid4().hex
        else:
            try:
                key = record[key_field]  # type: ignore
            except KeyError:
                raise ValueError(f"Key field {key_field} not found in record {record}")
        return f"{self._prefix}:{key}"

    @check_connected("_redis_conn")
    def info(self) -> Dict[str, Any]:
        """Get information about the index.

        Returns:
            dict: A dictionary containing the information about the index.
        """
        return convert_bytes(self._redis_conn.ft(self._name).info())  # type: ignore

    def create(self, overwrite: Optional[bool] = False):
        """Create an index in Redis from this SearchIndex object.

        Args:
            overwrite (bool, optional): Overwrite the index if it already exists. Defaults to False.

        Raises:
            redis.exceptions.ResponseError: If the index already exists.
        """
        raise NotImplementedError

    def delete(self, drop: bool = True):
        """Delete the search index.

        Args:
            drop (bool, optional): Delete the documents in the index. Defaults to True.

        raises:
            redis.exceptions.ResponseError: If the index does not exist.
        """
        raise NotImplementedError

    def load(
        self, data: Iterable[Dict[str, Any]], key_field: Optional[str] = None, **kwargs
    ):
        """Load data into Redis and index using this SearchIndex object.

        Args:
            data (Iterable[Dict[str, Any]]): An iterable of dictionaries
                containing the data to be indexed.
            key_field (Optional[str], optional): A field within the record
                to use in the Redis hash key.

        raises:
            redis.exceptions.ResponseError: If the index does not exist.
        """
        raise NotImplementedError


class SearchIndex(SearchIndexBase):
    """A class for interacting with Redis as a vector database.

    This class is a wrapper around the redis-py client that provides
    purpose-built methods for interacting with Redis as a vector database.

    Example:
        >>> from redisvl.index import SearchIndex
        >>> index = SearchIndex.from_yaml("schema.yaml")
        >>> index.create(overwrite=True)
        >>> index.load(data) # data is an iterable of dictionaries
    """

    def __init__(
        self,
        name: str,
        prefix: str = "rvl",
        storage_type: Optional[str] = "hash",
        fields: Optional[List["Field"]] = None,
    ):
        super().__init__(name, prefix, storage_type, fields)

    @classmethod
    def from_existing(
        cls,
        name: str,
        url: Optional[str] = None,
        fields: Optional[List["Field"]] = None,
        **kwargs,
    ):
        """Create a SearchIndex from an existing index in Redis.

        Args:
            name (str): Index name.
            url (Optional[str], optional): Redis URL. REDIS_URL env var
                is used if not provided. Defaults to None.
            fields (Optional[List[Field]], optional): List of Redis search
                fields to include in the schema. Defaults to None.

        Returns:
            SearchIndex: A SearchIndex object.

        Raises:
            redis.exceptions.ResponseError: If the index does not exist.
            ValueError: If the REDIS_URL env var is not set and url is not provided.

        """
        client = get_redis_connection(url, **kwargs)
        info = convert_bytes(client.ft(name).info())
        index_definition = make_dict(info["index_definition"])
        storage_type = index_definition["key_type"].lower()
        prefix = index_definition["prefixes"][0]
        instance = cls(
            name=name,
            storage_type=storage_type,
            prefix=prefix,
            fields=fields,
        )
        instance.set_client(client)
        return instance

    def connect(self, url: Optional[str] = None, **kwargs):
        """Connect to a Redis instance.

        Args:
            url (str): Redis URL. REDIS_URL env var is used if not provided.

        Raises:
            redis.exceptions.ConnectionError: If the connection to Redis fails.
            ValueError: If the REDIS_URL env var is not set and url is not provided.
        """
        self._redis_conn = get_redis_connection(url, **kwargs)

    @check_connected("_redis_conn")
    def create(self, overwrite: Optional[bool] = False):
        """Create an index in Redis from this SearchIndex object.

        Args:
            overwrite (bool, optional): Overwrite the index if it already exists. Defaults to False.

        Raises:
            redis.exceptions.ResponseError: If the index already exists.
        """
        check_redis_modules_exist(self._redis_conn)

        if not self._fields:
            raise ValueError("No fields defined for index")

        if self.exists() and overwrite:
            self.delete()

        # set storage_type, default to hash
        storage_type = IndexType.HASH
        if self._storage.lower() == "json":
            self._storage = IndexType.JSON

        # Create Index
        # will raise correct response error if index already exists
        self._redis_conn.ft(self._name).create_index(  # type: ignore
            fields=self._fields,
            definition=IndexDefinition(prefix=[self._prefix], index_type=storage_type),
        )

    @check_connected("_redis_conn")
    def delete(self, drop: bool = True):
        """Delete the search index.

        Args:
            drop (bool, optional): Delete the documents in the index. Defaults to True.

        raises:
            redis.exceptions.ResponseError: If the index does not exist.
        """
        # Delete the search index
        self._redis_conn.ft(self._name).dropindex(delete_documents=drop)  # type: ignore

    @check_connected("_redis_conn")
    def load(
        self, data: Iterable[Dict[str, Any]], key_field: Optional[str] = None, **kwargs
    ):
        """Load data into Redis and index using this SearchIndex object.

        Args:
            data (Iterable[Dict[str, Any]]): An iterable of dictionaries
                containing the data to be indexed.
            key_field (Optional[str], optional): A field within the record to
                use in the Redis hash key.

        raises:
            redis.exceptions.ResponseError: If the index does not exist.
        """
        # TODO -- should we return a count of the upserts? or some kind of metadata?
        if data:
            if not isinstance(data, Iterable):
                if not isinstance(data[0], dict):
                    raise TypeError("data must be an iterable of dictionaries")

            # Check if outer interface passes in TTL on load
            ttl = kwargs.get("ttl")
            with self._redis_conn.pipeline(transaction=False) as pipe:
                for record in data:
                    key = self._get_key(record, key_field)
                    pipe.hset(key, mapping=record)  # type: ignore
                    if ttl:
                        pipe.expire(key, ttl)
                pipe.execute()

    @check_connected("_redis_conn")
    def exists(self) -> bool:
        """Check if the index exists in Redis.

        Returns:
            bool: True if the index exists, False otherwise.
        """
        indices = convert_bytes(self._redis_conn.execute_command("FT._LIST"))  # type: ignore
        return self._name in indices


class AsyncSearchIndex(SearchIndexBase):
    """A class for interacting with Redis as a vector database asynchronously.

    This class is a wrapper around the redis-py client that provides
    purpose-built methods for interacting with Redis as a vector database.

    Example:
        >>> from redisvl.index import AsyncSearchIndex
        >>> index = AsyncSearchIndex.from_yaml("schema.yaml")
        >>> index.create(overwrite=True)
        >>> index.load(data) # data is an iterable of dictionaries
    """

    def __init__(
        self,
        name: str,
        prefix: str = "rvl",
        storage_type: Optional[str] = "hash",
        fields: Optional[List["Field"]] = None,
    ):
        super().__init__(name, prefix, storage_type, fields)

    @classmethod
    async def from_existing(
        cls,
        name: str,
        url: Optional[str] = None,
        fields: Optional[List["Field"]] = None,
        **kwargs,
    ):
        """Create a SearchIndex from an existing index in Redis.

        Args:
            name (str): Index name.
            url (Optional[str], optional): Redis URL. REDIS_URL env var
                is used if not provided. Defaults to None.
            fields (Optional[List[Field]], optional): List of Redis search
                fields to include in the schema. Defaults to None.

        Returns:
            SearchIndex: A SearchIndex object.

        Raises:
            redis.exceptions.ResponseError: If the index does not exist.
            ValueError: If the REDIS_URL env var is not set and url is not provided.

        """
        client = get_async_redis_connection(url, **kwargs)
        info = convert_bytes(await client.ft(name).info())
        index_definition = make_dict(info["index_definition"])
        storage_type = index_definition["key_type"].lower()
        prefix = index_definition["prefixes"][0]
        instance = cls(
            name=name,
            storage_type=storage_type,
            prefix=prefix,
            fields=fields,
        )
        instance.set_client(client)
        return instance

    def connect(self, url: Optional[str] = None, **kwargs):
        """Connect to a Redis instance.

        Args:
            url (str): Redis URL. REDIS_URL env var is used if not provided.

        Raises:
            redis.exceptions.ConnectionError: If the connection to Redis fails.
            ValueError: If no Redis URL is provided and REDIS_URL env var is not set.
        """
        self._redis_conn = get_async_redis_connection(url, **kwargs)

    @check_connected("_redis_conn")
    async def create(self, overwrite: Optional[bool] = False):
        """Create an index in Redis from this SearchIndex object.

        Args:
            overwrite (bool, optional): Overwrite the index if it already exists. Defaults to False.

        Raises:
            redis.exceptions.ResponseError: If the index already exists.
        """
        exists = await self.exists()
        if exists and overwrite:
            await self.delete()

        # set storage_type, default to hash
        storage_type = IndexType.HASH
        if self._storage.lower() == "json":
            self._storage = IndexType.JSON

        # Create Index
        await self._redis_conn.ft(self._name).create_index(  # type: ignore
            fields=self._fields,
            definition=IndexDefinition(prefix=[self._prefix], index_type=storage_type),
        )

    @check_connected("_redis_conn")
    async def delete(self, drop: bool = True):
        """Delete the search index.

        Args:
            drop (bool, optional): Delete the documents in the index. Defaults to True.

        raises:
            redis.exceptions.ResponseError: If the index does not exist.
        """
        # Delete the search index
        await self._redis_conn.ft(self._name).dropindex(delete_documents=drop)  # type: ignore

    @check_connected("_redis_conn")
    async def load(
        self,
        data: Iterable[Dict[str, Any]],
        concurrency: int = 10,
        key_field: Optional[str] = None,
        **kwargs,
    ):
        """Load data into Redis and index using this SearchIndex object.

        Args:
            data (Iterable[Dict[str, Any]]): An iterable of dictionaries
                containing the data to be indexed.
            concurrency (int, optional): Number of concurrent tasks to run. Defaults to 10.
            key_field (Optional[str], optional): A field within the record to
                use in the Redis hash key.

        raises:
            redis.exceptions.ResponseError: If the index does not exist.
        """
        ttl = kwargs.get("ttl")
        semaphore = asyncio.Semaphore(concurrency)

        async def _load(record: dict):
            async with semaphore:
                key = self._get_key(record, key_field)
                await self._redis_conn.hset(key, mapping=record)  # type: ignore
                if ttl:
                    await self._redis_conn.expire(key, ttl)

        # gather with concurrency
        await asyncio.gather(*[_load(record) for record in data])

    @check_connected("_redis_conn")
    async def search(self, *args, **kwargs) -> List["Result"]:
        """Perform a search on this index.

        Wrapper around redis.search.Search that adds the index name
        to the search query and passes along the rest of the arguments
        to the redis-py ft.search() method.

        Returns:
            List[Result]: A list of search results.
        """
        results: List["Result"] = await self._redis_conn.ft(self._name).search(*args, **kwargs)  # type: ignore
        return results

    async def query(self, query: "BaseQuery") -> List[Dict[str, Any]]:
        """Run a query on this index.

        This is similar to the search method, but takes a BaseQuery
        object directly (does not allow for the usage of a raw
        redis query string) and post-processes results of the search.

        Args:
            query (BaseQuery): The query to run.

        Returns:
            List[Result]: A list of search results.
        """
        results = await self.search(query.query, query_params=query.params)
        return process_results(results)

    @check_connected("_redis_conn")
    async def exists(self) -> bool:
        """Check if the index exists in Redis.

        Returns:
            bool: True if the index exists, False otherwise.
        """
        indices = await self._redis_conn.execute_command("FT._LIST")  # type: ignore
        return self._name in convert_bytes(indices)
